##!/usr/bin/perl -s


$usage = "
usage: $0 ham_count spam_count WordList < TestCorpus 

   -d         discounting parameter
   -h         this help
   -hh        this help

Read a unigram count file for ham and spam training as an argument.
Then process the test corpus and classify an email.  

Give the discounting parameter using -d=... . A value of 0.7 is
recommended.

";

die "$usage" if ( defined $h || defined $hh );

# set default value for delta
if(!defined($d)){
    $d=0.7;
}

$HamList  = shift;
$SpamList = shift;
$WordList = shift;

%vocOc = ();
open(WL,$WordList);
while($line = <WL>){
    chomp($line);
    $vocOc{$line} = 1;
}
close(WL);

die "counts required. See also: \n $usage" if (!(defined $HamList && defined $SpamList));


### MAIN ###


# read the unigram count files produced by corp2unicnt.pl

# First: read the ham count file
open (HamList) or die "Error opening $HamList: $!";

$NVocHam     = 0;
$NTrCountHam = 0;
$RHam        = 0;

while ( <HamList> ) {
    chomp;
    ($ti, $twHam, $tcHam) = split;
    if ( $tcHam > 0 ) {
	$RHam++;
    }
    $WListHam{$twHam} = $tcHam;
    $NVocHam++;
    $NTrCountHam += $tcHam;
}


# Second: read the spam count file
open (SpamList) or die "Error opening $SpamList: $!";

$NVocSpam     = 0;
$NTrCountSpam = 0;
$RSpam        = 0;

while ( <SpamList> ) {
    chomp;
    ($ti, $twSpam, $tcSpam) = split;
    if ( $tcSpam > 0 ) {
	$RSpam++;
    }
    $WListSpam{$twSpam} = $tcSpam;
    $NVocSpam++;
    $NTrCountSpam += $tcSpam;
}

$NWords = 0;
$LLSpam = 0.0; # spam likelihood
$LLHam  = 0.0; # ham likelihood


$correct = 0;
$false_neg = 0;
$false_pos = 0;
$incorrect = 0;


$testInstCounter = 0;
# loop over test corpus and over all emails. Classify the email
$isFirstLine = 1;
while ( <> ) {
    chomp;
    
    # header?
    if ($_ =~ /^\#\*\#\*\#/) {
	$testInstCounter++;
	if($isFirstLine){
	    # do nothing since we are in the first line
	    # and no words from a test instance have already been observed
	    $isFirstLine = 0;
	} else {
	    # is it the first header? (first header has 0 likelihood)
	    if (( $LLHam != 0.0 ) && ( $LLSpam != 0.0)) {
		#print "LL_Spam: $LLSpam\tLL_Ham: $LLHam\n";
		# get the perplexities
		$PPHam  = exp( $LLHam/$NWords ) ;
		$PPSpam = exp( $LLSpam/$NWords ) ;
	    
		print "PP_Spam: $PPSpam\tPP_Ham: $PPHam\n";
		
		# classify email
		if ($PPHam < $PPSpam) {     # if the perplexity of ham is smaller than spam
		    $newClass = "ham";       # then it is ham
		} else {
		    $newClass = "spam";      # else it it spam
		}
		
		# do statistics
		if ($class eq $newClass) {   # correct, if both classes are equal
		    $correct++;
		} else{
		    $incorrect++;
		    if ($class eq "ham") {
			$false_neg++;            # false negative, if ham is classified as spam
		    } else {
			$false_pos++;            # false positive, if spam is classified as ham
		    }
		}	    
		
		# reset some values
		$PPHam  = 0;
		$PPSpam = 0;
		$LLHam  = 0;
		$LLSpam = 0;
		$NWords = 0;
	    } else {
		# we only observed words in the test document that were not in the vocabulary
		# so we cannot classify
		$incorrect++;
	    }
	    
	    # get the original class
	    ($dummy, $class) = split;
	}
    } else {
	
	if($vocOc{$_}){
	    
	    $NWords++;
	    
	    
	    
	    ### calc prob for ham: ###
	    
	    $probHam = $d * $RHam / $NTrCountHam / $NVocHam;
	    # this term is common for seen and unseen words
	    
	    if ( (defined $WListHam{$_}) && ($WListHam{$_} > 0) ) {    # is word seen in training?
		$probHam += ($WListHam{$_} - $d ) / $NTrCountHam;
	    }
	    
	    $scoreHam = -log($probHam);
	    $LLHam += $scoreHam;
	    
	    
	    ### calc prob for spam: ###
	    
	    $probSpam = $d * $RSpam / $NTrCountSpam / $NVocSpam;       # this term is common for seen and unseen words
	    
	    if ( (defined $WListSpam{$_}) && ($WListSpam{$_} > 0) ) {  # is word seen in training?
		$probSpam += ($WListSpam{$_} - $d ) / $NTrCountSpam;
	    }


	    $scoreSpam = -log($probSpam);
	    $LLSpam += $scoreSpam;
	}
    }
}


# get the perplexities for the last email

$PPHam = exp( $LLHam/$NWords ) ;
$PPSpam = exp( $LLSpam/$NWords ) ;

# classify email
if ($PPHam < $PPSpam) {
    $newClass = "ham";
} else {
    $newClass = "spam";
}


if ($class eq $newClass) {
    $correct++;
} else{
    $incorrect++;
    if ($class eq "ham") {
	$false_neg++;
    } else {
	$false_pos++;
    }
}

# print statistics
print "Number of correctly classified test instances: $correct\nNumber of correctly classified test instances: $incorrect\nNumber of false positives: $false_pos\nNumber of false negatives: $false_neg\n\n";
print "Number of test instances: $testInstCounter\n";
